// Visit QWERTZ-tek at http://qwertz.w10.site/ for my other stuff
//
// Copyright (C) 2025 T. Sosnow. <t-sosnow@posteo.com>
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar.
// See the COPYING file or http://www.wtfpl.net/ for more details.
//

#include <stdio.h>
#include <conio.h>
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

// Some code for serial port handling has been stolen from:
// http://web.archive.org/web/20111027074853/http://www.robbayer.com/files/serial-win.pdf

HANDLE hSerial;
DWORD dwBytesRead = 0;
DWORD dwBytesWrote = 0;
void End(void);
unsigned char readPort();
void writePort(unsigned char szBuff);
int waitForCMD();
int waitForSomething();
void writeStringPort(unsigned char string[]);

char port[6];
DCB dcbSerialParams = {0};
COMMTIMEOUTS timeouts = {0};

unsigned char romfile[65536];

int main(int argc, char *argv[]) {
    unsigned char name[256]={0};
    unsigned char ROM_name[38][11]= {"2716","X2816A","DQ52B13","2816A","2817A","2732","2732A","TMS2532","MCM2532","DQ52B23","MCM2833",
                            "2764","MBM2764","2764A","AM2764A","P2764A","TMS2564","MCM68764","RD687C64","LH5764","X2864A",
                            "DQ52B33","HN58064","27128","MBM27128","27128A","AM27128A","P27128","27916","LH57128","27256",
                            "AM27256","P27256","MBM27C256","LH57256","27512","AM27512","27513"
                           };
    unsigned char ROM_size[38]= {0,0,0,0,0,1,1,1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,4,4,4,4,4,5,5,5};
    unsigned char ROM_volt[38]= {5,0,0,0,0,5,4,5,5,0,0,4,4,1,3,2,5,5,5,1,0,0,0,5,5,1,3,2,4,1,1,4,2,4,1,1,3,1};
    unsigned char ROM_type[38][5]= {"16","216A","5213","816A","817A","32","32A","532","632","5223","2833","64","F64","64A","A64A",
                           "B64A","564","664","D64","5764","264A","5233","8064","128","F28","28A","A28A","B28A","916",
                           "5728","256","A56","B56","F56","5756","512","A512","513"
                          };
    unsigned char voltages[6][7]= {" ","12.5V","12.75V","13.0V","21.0V","25.0V"};
    int i,a,b,c;
    char e[3];
    int fsize;
    unsigned int short chksum=0;
    unsigned int long recsum;
    unsigned char portbyte;
    int tocounts;
//    unsigned char serialtemp[33];
    FILE *romf;
    printf("SEFTrans V0.4\nTransfer tool for Advantest SE4943 (1985), aka TR4943\nBy T. Sosnow. 2015-2017\n\n");

    strncpy (&port[0], "COM1",6);

    hSerial = CreateFile(port,
                         GENERIC_READ | GENERIC_WRITE,
                         0,
                         0,
                         OPEN_EXISTING,
                         FILE_ATTRIBUTE_NORMAL,
                         0);
    if(hSerial==INVALID_HANDLE_VALUE) {
        if(GetLastError()==ERROR_FILE_NOT_FOUND) {
            printf("Port %s doesn't exist\n",port);
            return 1;
        }
        printf("Can't access Port %s\n",port);
        return 2;
    }
    atexit(End);
    //DCB dcbSerialParams = {0};
    dcbSerialParams.DCBlength=sizeof(dcbSerialParams);
    if (!GetCommState(hSerial, &dcbSerialParams)) {
        printf("Can't get Port %s state\n",port);
        return 3;
    }
    dcbSerialParams.BaudRate=CBR_19200;
    dcbSerialParams.ByteSize=8;
    dcbSerialParams.StopBits=ONESTOPBIT; //ONESTOPBIT, ONE5STOPBITS, TWOSTOPBITS
    dcbSerialParams.Parity=NOPARITY; //EVENPARITY, NOPARITY, ODDPARITY
    dcbSerialParams.fOutxCtsFlow=1;
    dcbSerialParams.fRtsControl=RTS_CONTROL_HANDSHAKE;
    if(!SetCommState(hSerial, &dcbSerialParams)) {
        printf("Can't set Port %s state\n",port);
        return 4;
    }
    timeouts.ReadIntervalTimeout=50;
    timeouts.ReadTotalTimeoutConstant=50;
    timeouts.ReadTotalTimeoutMultiplier=10;
    timeouts.WriteTotalTimeoutConstant=50;
    timeouts.WriteTotalTimeoutMultiplier=10;
    if(!SetCommTimeouts(hSerial, &timeouts)) {
        printf("Can't set Port %s timeouts\n",port);
        return 5;
    }
    writePort(0x1B);
    Sleep(50);
    writePort(0x11);
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("S900000 01\r");
waitMenu: ;
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    };
//    printf("\nPush Select then A, set '19200, NON, XOF'\n");
//    printf("Push Select then 9, set 'DG-BIN NO, NON'\n\n");
//    if(argc>1) goto processArgs;
printMenu: ;
    printf("\nSelect option:\n1. File->Buffer\n2. Buffer->File\n3. List ROM types\n4. Set ROM type\n5. ROM->Buffer (Dump)\n6. Compare ROM with Buffer\n7. Buffer->ROM (Program)\n8. Erase ROM (if supported)\n9. Blank check ROM\n(Q)uit\n");
menu: ;
    fgets(e,2,stdin);
    if (e[0]=='\n') {
        goto menu;
    }
    if (e[0]=='1') {
        goto readfromfile;
    } else if (e[0]=='2') {
        goto readtofile;
    } else if (e[0]=='3') {
        goto listroms;
    } else if (e[0]=='4') {
        goto setrom;
    } else if (e[0]=='5') {
        goto readrom;
    } else if (e[0]=='6') {
        goto comparerom;
    } else if (e[0]=='7') {
        goto programrom;
    } else if (e[0]=='8') {
        goto eraserom;
    } else if (e[0]=='9') {
        goto blankchk;
    } else if (e[0]=='q'||e[0]=='Q') {
        return 0;
    } else {
        printf("Option %s does not exist, please try again.\n",e);
        goto menu;
    }

listroms: ;
    i=0;
    a=0;
    printf("\nROM Chip   Volt.  Type | ROM Chip   Volt.  Type | ROM Chip   Volt.  Type |\n");
    printf("---------------------------  2 kilobyte ROMs ---------------------------\n");
    while(i<38) {
        printf("%-10s %-6s %-4s | ",ROM_name[i],voltages[ROM_volt[i]],ROM_type[i]);
        i++;
        a++;
        if(a%3==0) printf("\n");
        if(ROM_size[i]!=ROM_size[i-1]) {
            if(a%3!=0) printf("\n");
            if(ROM_size[i]==1) printf("---------------------------  4 kilobyte ROMs ---------------------------\n");
            if(ROM_size[i]==2) printf("---------------------------  8 kilobyte ROMs ---------------------------\n");
            if(ROM_size[i]==3) printf("--------------------------- 16 kilobyte ROMs ---------------------------\n");
            if(ROM_size[i]==4) printf("--------------------------- 32 kilobyte ROMs ---------------------------\n");
            if(ROM_size[i]==5) printf("--------------------------- 64 kilobyte ROMs ---------------------------\n");
            a=0;
        }
    }
    printf("\nPress any key to continue\n");
    fflush(stdin);
    while(!kbhit());
    fflush(stdin);
    goto printMenu;


setrom: ;
    fflush(stdin);
    printf("Enter ROM type (full or short name)\n");
    scanf("%s",name);
    i=0;
    while(i<38) {
        if(strcasecmp(name,ROM_name[i])==0) goto sendromtype;
        i++;
    }
    i=0;
    while(i<38) {
        if(strcasecmp(name,ROM_type[i])==0) goto sendromtype;
        i++;
    }
    printf("Unknown ROM type, see ROM list for supported types.\n");
    goto printMenu;
sendromtype: ;
    printf("Type:%s Voltage:%s Short name:%s\n",ROM_name[i],voltages[ROM_volt[i]],ROM_type[i]);
    writePort('R');
    writeStringPort(ROM_type[i]);
    writePort('\r');
    fflush(stdin);
    goto printMenu;


readrom: ;
    writeStringPort("R\r");
    Sleep(100);
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]='\0';
//    printf("read response:%s\n",name);
    while(name[0]=='0'){
        name[0]=name[1];
        name[1]=name[2];
        name[2]=name[3];
        name[3]=name[4];
        name[4]='\0';
    }
//    printf("padded response:%s\n",name);
    i=0;
    while(i<38) {
        if(strcasecmp(name,ROM_type[i])==0) break;
        i++;
    }
    switch(ROM_size[i]) {
    case 0 : fsize=0x800;
        break;
    case 1 : fsize=0x1000;
        break;
    case 2 : fsize=0x2000;
        break;
    case 3 : fsize=0x4000;
        break;
    case 4 : fsize=0x8000;
        break;
    case 5 : fsize=0x10000;
    }
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("P8\r");
    Sleep(100);
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("P\r");
    Sleep((fsize/0x100)+1000);
    if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]=readPort();
    if(strcasecmp(name,"PASS]")!=0) {
        if(strcasecmp(name,"ERROR")==0) {
            name[0]=readPort();
            name[1]=readPort();
            name[2]=readPort();
            name[3]=readPort();
            name[4]=readPort();
            if(strcasecmp(name," #06]")==0) {
                printf("Operation failed!\nCheck chip contacts!\n");
                fflush(stdin);
                goto waitMenu;
            }else if(strcasecmp(name," #10]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #11]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #12]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #13]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #14]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #15]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #16]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #17]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else{
                printf("Unexpected error #%c%c\n",name[2],name[3]);
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }
        }
        printf("Invalid Response\n");
        return 7;
    }
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]=readPort();
    if(strcasecmp(name,"PASS]")!=0) {
        if(strcasecmp(name,"ERROR")==0) {
            name[0]=readPort();
            name[1]=readPort();
            name[2]=readPort();
            name[3]=readPort();
            name[4]=readPort();
            if(strcasecmp(name," #06]")==0) {
                printf("Operation failed!\nCheck chip contacts!\n");
                fflush(stdin);
                goto waitMenu;
            }else if(strcasecmp(name," #10]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #11]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #12]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #13]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #14]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #15]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #16]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #17]")==0){
                printf("Operation failed!\nVerify error!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else{
                printf("Unexpected error #%c%c\n",name[2],name[3]);
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }
        }
        printf("Invalid Response\n");
        return 7;
    }
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    fflush(stdin);
    goto printMenu;

comparerom: ;
    writeStringPort("R\r");
    Sleep(100);
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]='\0';
//    printf("read response:%s\n",name);
    while(name[0]=='0'){
        name[0]=name[1];
        name[1]=name[2];
        name[2]=name[3];
        name[3]=name[4];
        name[4]='\0';
    }
//    printf("padded response:%s\n",name);
    i=0;
    while(i<38) {
        if(strcasecmp(name,ROM_type[i])==0) break;
        i++;
    }
    switch(ROM_size[i]) {
    case 0 : fsize=0x800;
        break;
    case 1 : fsize=0x1000;
        break;
    case 2 : fsize=0x2000;
        break;
    case 3 : fsize=0x4000;
        break;
    case 4 : fsize=0x8000;
        break;
    case 5 : fsize=0x10000;
    }
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("PE\r");
    Sleep(100);
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("P\r");
    Sleep((fsize/0x4000)+500);
    if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]=readPort();
    name[5]='\0';
    if(strcasecmp(name,"PASS]")!=0) {
        if(strcasecmp(name,"ERROR")==0) {
            name[0]=readPort();
            name[1]=readPort();
            name[2]=readPort();
            name[3]=readPort();
            name[4]=readPort();
            if(strcasecmp(name," #06]")==0) {
                printf("Operation failed!\nCheck chip contacts!\n");
                fflush(stdin);
                goto waitMenu;
            }else if(strcasecmp(name," #10]")==0){
                printf("ROM doesn't match Buffer!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #11]")==0){
                printf("ROM doesn't match Buffer!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #12]")==0){
                printf("ROM doesn't match Buffer!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #13]")==0){
                printf("ROM doesn't match Buffer!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #14]")==0){
                printf("ROM doesn't match Buffer!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #15]")==0){
                printf("ROM doesn't match Buffer!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #16]")==0){
                printf("ROM doesn't match Buffer!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #17]")==0){
                printf("ROM doesn't match Buffer!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else{
                printf("Unexpected error #%c%c\n",name[2],name[3]);
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }
        }
        printf("Invalid Response\n");
        return 7;
    }
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    fflush(stdin);
    goto printMenu;

programrom: ;
    writeStringPort("R\r");
    Sleep(100);
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]='\0';
//    printf("read response:%s\n",name);
    while(name[0]=='0'){
        name[0]=name[1];
        name[1]=name[2];
        name[2]=name[3];
        name[3]=name[4];
        name[4]='\0';
    }
//    printf("padded response:%s\n",name);
    i=0;
    while(i<38) {
        if(strcasecmp(name,ROM_type[i])==0) break;
        i++;
    }
    switch(ROM_size[i]) {
    case 0 : fsize=0x800;
        break;
    case 1 : fsize=0x1000;
        break;
    case 2 : fsize=0x2000;
        break;
    case 3 : fsize=0x4000;
        break;
    case 4 : fsize=0x8000;
        break;
    case 5 : fsize=0x10000;
    }
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    
    printf("Type the word 'PROGRAM' to confirm this action\n");
    scanf("%s",name);
    if(strcasecmp(name,"PROGRAM")!=0){
        goto printMenu;
    }
     writeStringPort("PD\r");
    Sleep(100);
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("P\r");
    Sleep(fsize*5);
    if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]=readPort();
    name[5]='\0';
    if(strcasecmp(name,"PASS]")!=0) {
        if(strcasecmp(name,"ERROR")==0) {
            name[0]=readPort();
            name[1]=readPort();
            name[2]=readPort();
            name[3]=readPort();
            name[4]=readPort();
            if(strcasecmp(name," #06]")==0) {
                printf("Operation failed!\nCheck chip contacts!\n");
                fflush(stdin);
                goto waitMenu;
            }else if(strcasecmp(name," #10]")==0){
                printf("ROM couldn't be programmed!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #11]")==0){
                printf("ROM couldn't be programmed!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #12]")==0){
                printf("ROM couldn't be programmed!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #13]")==0){
                printf("ROM couldn't be programmed!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #14]")==0){
                printf("ROM couldn't be programmed!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #15]")==0){
                printf("ROM couldn't be programmed!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #16]")==0){
                printf("ROM couldn't be programmed!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #17]")==0){
                printf("ROM couldn't be programmed!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #05]")==0){
                printf("ROM error!\nReinsert ROM\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else{
                printf("Unexpected error #%c%c\n",name[2],name[3]);
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }
        }
        printf("Invalid Response\n");
        return 7;
    }
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    fflush(stdin);
    goto printMenu;

eraserom: ;
    printf("Type the word 'ERASE' to confirm this action\n");
    scanf("%s",name);
    if(strcasecmp(name,"erase")!=0){
        goto printMenu;
    }
    writeStringPort("P9\r");
    Sleep(100);
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("P\r");
    if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]=readPort();
    name[5]='\0';
    if(strcasecmp(name,"PASS]")!=0) {
        if(strcasecmp(name,"ERROR")==0) {
            name[0]=readPort();
            name[1]=readPort();
            name[2]=readPort();
            name[3]=readPort();
            name[4]=readPort();
            if(strcasecmp(name," #06]")==0) {
                printf("Operation failed!\nCheck chip contacts!\n");
                fflush(stdin);
                goto waitMenu;
            }else if(strcasecmp(name," #10]")==0){
                printf("ROM couldn't be erased!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #11]")==0){
                printf("ROM couldn't be erased!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #12]")==0){
                printf("ROM couldn't be erased!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #13]")==0){
                printf("ROM couldn't be erased!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #14]")==0){
                printf("ROM couldn't be erased!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #15]")==0){
                printf("ROM couldn't be erased!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #16]")==0){
                printf("ROM couldn't be erased!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #17]")==0){
                printf("ROM couldn't be erased!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #35]")==0){
                printf("Only EEPROMs can be erased with this function!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else{
                printf("Unexpected error #%c%c\n",name[2],name[3]);
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }
        }
        printf("Invalid Response\n");
        return 7;
    }
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    fflush(stdin);
    goto printMenu;



blankchk: ;
    writeStringPort("R\r");
    Sleep(100);
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]='\0';
//    printf("read response:%s\n",name);
    while(name[0]=='0'){
        name[0]=name[1];
        name[1]=name[2];
        name[2]=name[3];
        name[3]=name[4];
        name[4]='\0';
    }
//    printf("padded response:%s\n",name);
    i=0;
    while(i<38) {
        if(strcasecmp(name,ROM_type[i])==0) break;
        i++;
    }
    switch(ROM_size[i]) {
    case 0 : fsize=0x800;
        break;
    case 1 : fsize=0x1000;
        break;
    case 2 : fsize=0x2000;
        break;
    case 3 : fsize=0x4000;
        break;
    case 4 : fsize=0x8000;
        break;
    case 5 : fsize=0x10000;
    }
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("PC\r");
    Sleep(100);
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    writeStringPort("P\r");
    Sleep((fsize/0x4000)+500);
    if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]=readPort();
    name[5]='\0';
    if(strcasecmp(name,"PASS]")!=0) {
        if(strcasecmp(name,"ERROR")==0) {
            name[0]=readPort();
            name[1]=readPort();
            name[2]=readPort();
            name[3]=readPort();
            name[4]=readPort();
            if(strcasecmp(name," #06]")==0) {
                printf("Operation failed!\nCheck chip contacts!\n");
                fflush(stdin);
                goto waitMenu;
            }else if(strcasecmp(name," #10]")==0){
                printf("ROM not blank!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #11]")==0){
                printf("ROM not blank!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #12]")==0){
                printf("ROM not blank!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #13]")==0){
                printf("ROM not blank!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #14]")==0){
                printf("ROM not blank!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #15]")==0){
                printf("ROM not blank!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #16]")==0){
                printf("ROM not blank!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else if(strcasecmp(name," #17]")==0){
                printf("ROM not blank!\n");
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }else{
                printf("Unexpected error #%c%c\n",name[2],name[3]);
                writePort(0x1B);
                Sleep(200);
                goto waitMenu;
            }
        }
        printf("Invalid Response\n");
        return 7;
    }
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    fflush(stdin);
    goto printMenu;


readfromfile: ;

typerfile: ;
    printf("Enter Filename\n");
//    scanf("%s",name);
    fflush(stdin);
    name[0]='\0';
    gets_s(name, sizeof(name));
    readthefile: ;
    if(name[0]=='"'){
        i=0;
        while(i<(sizeof(name)-1)){
            if(name[i+1]=='"') name[i+1]='\0';
            name[i]=name[i+1];
            i++;
        }
    }else if(name[0]=='\0'){
        fflush(stdin);
        goto printMenu;
    }
    romf = fopen (name, "rb");
    if (romf != NULL) {
        fseek (romf, 0, SEEK_END);
        fsize=ftell (romf);
        rewind (romf);
        if(!((fsize==0x800)||(fsize==0x1000)||(fsize==0x2000)||(fsize==0x4000)||(fsize==0x8000)||(fsize==0x10000))) {
            printf("Unsupported file size\n");
//            if(argc>1) return 6;
            goto typerfile;
        }
        i=0;
        while(i<fsize) {
            romfile[i]=fgetc(romf);
            i++;
        }
        fclose (romf);
    } else {
        printf("Can't read file %s\n",name);
        goto typerfile;
    }
//   printf("\n");
    //  printf("Push Select then 6 and start transfer\n\n");

    //  printf("Press any key to continue\n");
    //  fflush(stdin);
    //  getchar();

    writeStringPort("S6 0\r");
    Sleep(100);
//    printf("Now transferring data\n");
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0x00);
    writePort(0xFF);
    i=0;
    chksum=0;
    while(i<fsize) {
        writePort(romfile[i]);
        if(dwBytesWrote) {
            chksum+=romfile[i];
            i++;
        }
        if(kbhit()) {
            if(getch()==0x1B) {
                EscapeCommFunction(hSerial, SETBREAK);
                Sleep(500);
                EscapeCommFunction(hSerial, CLRBREAK);
                writePort(0x1B);
                Sleep(500);
                goto waitMenu;
            }
        }
    }
    Sleep(200);
//    writePort(0xff);
//    writePort(0x00);
//    writePort(0xff);
//    writePort(0x00);
//    printf("\nPush Select then 1, choose BLOCK\n");
//    printf("Verify checksum:%4X\n\n",chksum);
    EscapeCommFunction(hSerial, SETBREAK);
    Sleep(500);
    EscapeCommFunction(hSerial, CLRBREAK);
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }

    writeStringPort("S1K 0000-");
    switch(fsize) {
    case 0x800 : writeStringPort("07");
        break;
    case 0x1000 : writeStringPort("0F");
        break;
    case 0x2000 : writeStringPort("1F");
        break;
    case 0x4000 : writeStringPort("3F");
        break;
    case 0x8000 : writeStringPort("7F");
        break;
    case 0x10000 : writeStringPort("FF");
    }
    writeStringPort("FF\r");
    Sleep((fsize/0x4000)+700);
    if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]=readPort();
    name[5]=readPort();
    if(!(name[4]==' '&&name[5]==']')) {
        printf("Invalid Response\n");
        return 7;
    }
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    name[4]='\0';
    recsum=strtoul(name, NULL,16);
    if(recsum!=chksum) {
        printf("Checksum wrong! %04x does not equal %04x\n",recsum,chksum);
        return 7;
    }

    /*printf("\nPress any key to continue\n");
        fflush(stdin);
        getchar();*/
    goto printMenu;

readtofile: ;
    writeStringPort("R\r");
    Sleep(100);
        if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]='\0';
//    printf("read response:%s\n",name);
    while(name[0]=='0'){
        name[0]=name[1];
        name[1]=name[2];
        name[2]=name[3];
        name[3]=name[4];
        name[4]='\0';
    }
//    printf("padded response:%s\n",name);
    i=0;
    while(i<38) {
        if(strcasecmp(name,ROM_type[i])==0) break;
        i++;
    }
    switch(ROM_size[i]) {
    case 0 : fsize=0x800;
        break;
    case 1 : fsize=0x1000;
        break;
    case 2 : fsize=0x2000;
        break;
    case 3 : fsize=0x4000;
        break;
    case 4 : fsize=0x8000;
        break;
    case 5 : fsize=0x10000;
    }
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
//    printf("\n");
//    printf("Push Select then 8 and start transfer\n\n");
    writeStringPort("S8\r");
    i=0;
    fflush(stdin);
    while(1) {
        portbyte=readPort();
        if (portbyte==0x00&&dwBytesRead) i++;
        if (portbyte==0xFF&&dwBytesRead) break;
        if(kbhit()) {
            if(getch()==0x1B) {
                writePort(0x1B);
                Sleep(5000);
                goto waitMenu;
            }
        }
    }
    if(i>200||i<2) {
        printf("Data preamble invalid\n");
        return 6;
//    }else{
//    printf("Data preamble valid\n");
    }
    tocounts=0;
    i=0;
//    chksum=0;
    while(1) {
        portbyte=readPort();
        if(!dwBytesRead) tocounts++;
        if(tocounts>25) break;
        if(dwBytesRead) {
            romfile[i]=portbyte;
//            chksum+=portbyte;
            i++;
        }
        if(i==0x10000) break;
        if(kbhit()) {
            if(getch()==0x1B) {
                writePort(0x1B);
                Sleep(2000);
                goto waitMenu;
            }
        }
    }
    if(i==0x10000){
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
        a=i;
        goto skipCMDscan;
    }
    
    if(i!=fsize) {
        if((i-fsize)<250) {
            goto skipRecChk;
        }
        printf("Transmission has been aborted, %i bytes has been received\n",i);
        return 6;
    }



//    printf("Push Select then 1, choose BLOCK\n");
//    printf("Verify checksum:%4X\n\n",chksum);

//    fsize=i;

skipRecChk: ;
    a=i;
    if((romfile[i-5]==0x0D)&&(romfile[i-4]==0x0A)&&(romfile[i-3]=='[')&&(romfile[i-2]=='C')&&(romfile[i-1]==']')){
    a-=6;
    }else{
        printf("No Response\n");
        return 6;
    }
skipCMDscan: ;
    while(a>fsize) {
        if(romfile[a]!=0x00) {
            printf("Transmission has been aborted, %i bytes has been received\n",i);
        }
        a--;
    }
    i=0;
    chksum=0;
    while(i<fsize) {
        chksum+=romfile[i];
        i++;
    }

    writeStringPort("S1K 0000-");
    switch(fsize) {
    case 0x800 : writeStringPort("07");
        break;
    case 0x1000 : writeStringPort("0F");
        break;
    case 0x2000 : writeStringPort("1F");
        break;
    case 0x4000 : writeStringPort("3F");
        break;
    case 0x8000 : writeStringPort("7F");
        break;
    case 0x10000 : writeStringPort("FF");
    }
    writeStringPort("FF\r");
    Sleep((fsize/0x4000)+700);
    if(!waitForSomething()) {
        printf("No Response\n");
        return 6;
    }
    name[0]=readPort();
    name[1]=readPort();
    name[2]=readPort();
    name[3]=readPort();
    name[4]=readPort();
    name[5]=readPort();
    if(!(name[4]==' '&&name[5]==']')) {
        printf("Invalid Response\n");
        return 7;
    }
    if(!waitForCMD()) {
        printf("No Response\n");
        return 6;
    }
    name[4]='\0';
    recsum=strtoul(name, NULL,16);
    if(recsum!=chksum) {
        printf("Checksum wrong! %04x does not equal %04x\n",recsum,chksum);
        return 7;
  }

    i=0;

//    if(argc>1) goto getNameToWrite;


typefile: ;
    printf("Enter Filename\n");
//    scanf("%s",name);
    fflush(stdin);
    name[0]='\0';
    gets_s(name, sizeof(name));
writefile: ;
    if(name[0]=='"'){
        i=0;
        while(i<(sizeof(name)-1)){
            if(name[i+1]=='"') name[i+1]='\0';
            name[i]=name[i+1];
            i++;
        }
//    }else if(name[0]=='\0'){
//        fflush(stdin);
//        goto printMenu;
    }
    romf = fopen (name, "wb");
    if (romf != NULL) {
        while(i<fsize) {
            fputc(romfile[i], romf);
            i++;
        }

        fclose (romf);
    fflush(stdin);
    goto printMenu;
    } else {
        printf("Can't write file %s\n",name);
        goto typefile;
    }



    
/*
printcmdhelp:;
    printf("  -?  -  Shows this prompt.\n  -r file  -  Reads file and uploads it.\n  -w file  -  Downloads data and writes it to file.\n  -rom ROM_TYPE  -  Sets the ROM type, ? lists all types with their voltage\n  -copy  -  Copy the inserted ROM to RAM\n  -read  -  Compare the inserted ROM and RAM\n  -program  -  Program inserted ROM with RAM\n  -erase  -  Erase ROM (if supported)\n  -blank  -  Check if inserted ROM is blank\n");
    return 0;

processArgs:;
    if(!strncasecmp(argv[1], "-?", 2)) goto printcmdhelp;
    if(!strncasecmp(argv[1], "--help", 6)) goto printcmdhelp;
    if(argc!=3) {
        printf("Wrong number of arguments");
        return 0;
    }
    if(!strncasecmp(argv[1], "-w", 2)) goto readtofile;
    strcpy(name,argv[2]);
    if(!strncasecmp(argv[1], "-r", 2)) goto readthefile;

    printf("Unknown command: %s\n",argv[1]);
    return 255;


getNameToWrite: ;
    strcpy(name,argv[2]);
    goto writefile;
*/

//    printf("This is a placeholder error\nPlease try again\n");
//    return 420;

}

unsigned char readPort() {
    unsigned char szBuff = 0;
    if(!ReadFile(hSerial, &szBuff, 1, &dwBytesRead, NULL)) {
        printf("Couldn't read Port %s\n",port);
    }
    return szBuff;
}
void writeStringPort(unsigned char string[]) {
    int count=0;
    while(string[count]!='\0') {
        writePort(string[count]);
        count++;
    }
}

void writePort(unsigned char szBuff) {
    if(!WriteFile(hSerial, &szBuff, 1, &dwBytesWrote, NULL)) {
        printf("Couldn't write to Port %s\n",port);
    }
}

int waitForCMD() {
    int wait=0,result=0;
    while(wait<3600) {
        if(readPort()=='['&&dwBytesRead) {
            if(readPort()=='C'&&dwBytesRead) {
                if(readPort()==']'&&dwBytesRead) {
                    result++;
                    break;
                }
            }
        };
        wait++;
        Sleep(75);
    };
    return result;
};

int waitForSomething() {
    int wait=0,result=0;
    while(wait<3600) {
        if(readPort()=='['&&dwBytesRead) {
            result++;
            break;
        };
        wait++;
        Sleep(75);
    };
    return result;
};

void End(void) {
    writePort(0x1B);
    Sleep(100);
    writePort('S');
    writePort('C');
    writePort(0x0D);
    Sleep(100);
    CloseHandle(hSerial);
}
